#!/usr/bin/env python3

import re
from pathlib import Path

root = Path(__file__).parents[1]

version_py = root / 'src' / '_version.py'
version_str = version_py.read_text()
version_pattern = r"^__version__ = ['\"]([^'\"]*)['\"]"

match = re.search(version_pattern, version_str, re.M)
if match:
    print(match.group(1))
else:
    raise RuntimeError(f"Unable to find version string in: {version_py}")



import numpy as np
from py3nj import _wigner


def int_broadcast(*args):
    args = np.broadcast_arrays(*args)
    if any(a.dtype.kind not in "iu" for a in args):
        raise ValueError("Argument should be integer.")
    return args


def clebsch_gordan(
    two_j1, two_j2, two_j3, two_m1, two_m2, two_m3, ignore_invalid=False
):
    """ Calulate Clebsch-Gordan coefficient
    <j1 m1, j2 m2 | j3 m3>

    Parameters
    ----------
    two_j1: array of integers
    two_j2: array of integers
    two_j3: array of integers
    two_m1: array of integers
    two_m2: array of integers
    two_m3: array of integers
        Since j1, ..., m3 should be integers or half integers, two_j1 (which
        means 2 x j1) should be all integers.
    force_compute: boolean
        If True, returns 0 even for invalid arguments.
        Otherwise, raise a ValueError.

    Returns
    -------
    clebch-gordan: array
        The value of Clebsch Gordan coefficients, with the same size of the
        arguments.
    """
    two_j1, two_j2, two_j3, two_m1, two_m2, two_m3 = int_broadcast(
        two_j1, two_j2, two_j3, two_m1, two_m2, two_m3
    )

    phase = (two_j1 - two_j2 + two_m3) % 4 - 1
    return (
        -phase
        * np.sqrt(two_j3 + 1)
        * wigner3j(
            two_j1,
            two_j2,
            two_j3,
            two_m1,
            two_m2,
            -two_m3,
            ignore_invalid=ignore_invalid,
        )
    )


def wigner3j(two_l1, two_l2, two_l3, two_m1, two_m2, two_m3, ignore_invalid=False):
    """
    Calculate wigner 3j symbol
    (  L1   L2 L3)
    (-M2-M3 M2 M3)

    Parameters
    ----------
    two_l1: array of integers
    two_l2: array of integers
    two_l3: array of integers
    two_m1: array of integers
    two_m2: array of integers
    two_m3: array of integers
        Since L1, ..., M3 should be integers or half integers, two_l1 (which
        means 2 x L1) should be all integers.
    ignore_invalid: boolean
        If True, returns 0 even for invalid arguments.
        Otherwise, raise a ValueError.

    Returns
    -------
    threej: array
        The value of 3J symbol with the same shape of the arguments.
    """
    two_l1, two_l2, two_l3, two_m1, two_m2, two_m3 = int_broadcast(
        two_l1, two_l2, two_l3, two_m1, two_m2, two_m3
    )

    if (two_l1 < 0).any() and not force_compute:
        raise ValueError("Some of l values are negative")

    if ignore_invalid:
        valid_argument = two_l1 >= 0
        two_l2 = np.where(valid_argument, two_l2, 0)
        two_l3 = np.where(valid_argument, two_l3, 0)
        two_m2 = np.where(valid_argument, two_m2, 0)
        two_m3 = np.where(valid_argument, two_m3, 0)

    l, thrcof = drc3jj(two_l2, two_l3, two_m2, two_m3, ignore_invalid=ignore_invalid)

    l1max = thrcof.shape[-1]
    shape = thrcof.shape[:-1]

    two_l1 = two_l1.ravel()
    valid = (two_l1 < l1max) * ((two_m1 + two_m2 + two_m3).ravel() == 0)
    if ignore_invalid:
        valid = valid * valid_argument.ravel()

    # temporary set the invalid l1 to zero
    two_l1 = np.where(valid, two_l1, 0)
    if shape == ():  # scalar case
        return np.where(valid, thrcof[two_l1], 0.0)[0]
    thrcof = thrcof.reshape(-1, l1max)
    thrcof = np.where(valid, thrcof[np.arange(len(two_l1)), two_l1], 0.0)
    return thrcof.reshape(shape)


def wigner6j(two_l1, two_l2, two_l3, two_l4, two_l5, two_l6, ignore_invalid=False):
    """
    Calculate wigner 6j symbol
    (L1 L2 L3)
    (L4 L5 L6)

    Parameters
    ----------
    two_l1: array of integers
    two_l2: array of integers
    two_l3: array of integers
    two_l4: array of integers
    two_l5: array of integers
    two_l6: array of integers
        Since L1, ..., L6 should be integers or half integers, two_l1 (which
        means 2 x L1) should be all integers.
    ignore_invalid: boolean
        If True, returns 0 even for invalid arguments.
        Otherwise, raise a ValueError.

    Returns
    -------
    threej: array
        The value of 6J symbol with the same shape of the arguments.
    """
    two_l1, two_l2, two_l3, two_l4, two_l5, two_l6 = int_broadcast(
        two_l1, two_l2, two_l3, two_l4, two_l5, two_l6
    )

    if (two_l1 < 0).any() and not ignore_invalid:
        raise ValueError("Some of l values are negative")

    if ignore_invalid:
        valid_argument = two_l1 >= 0
        two_l1 = np.where(valid_argument, two_l1, 0)
        two_l2 = np.where(valid_argument, two_l2, 0)
        two_l3 = np.where(valid_argument, two_l3, 0)
        two_l4 = np.where(valid_argument, two_l4, 0)
        two_l5 = np.where(valid_argument, two_l5, 0)
        two_l6 = np.where(valid_argument, two_l6, 0)
        # TODO merge the valid condition in drc6j

    l, sixcof = drc6j(two_l2, two_l3, two_l4, two_l5, two_l6, ignore_invalid=ignore_invalid)

    l1max = sixcof.shape[-1]
    shape = sixcof.shape[:-1]

    two_l1 = two_l1.ravel()
    valid = two_l1 < l1max
    if ignore_invalid:
        valid = valid * valid_argument.ravel()

    # temporary set the invalid l1 to the first value
    two_l1 = np.where(valid, two_l1, 0)
    if shape == ():  # scale case
        return np.where(valid, sixcof[two_l1], 0.0)[0]
    sixcof = sixcof.reshape(-1, l1max)
    sixcof = np.where(valid, sixcof[np.arange(len(two_l1)), two_l1], 0.0)
    return sixcof.reshape(shape)


def wigner9j(two_l1, two_l2, two_l3, two_l4, two_l5, two_l6, two_l7, two_l8, two_l9):
    """
    Calculate wigner 9j symbol
    (L1 L2 L3)
    (L4 L5 L6)
    (L7 L8 L9)

    defined as
              2x        (a b c) (d e f) (g h j)
    sum_x (-1)   (2x+1) (f j x) (b x h) (x a d)

              2x        (x f b) (x h b) (x a j)
    sum_x (-1)   (2x+1) (c a j) (e f d) (g h d)


    Parameters
    ----------
    two_l1: array of integers
    two_l2: array of integers
    two_l3: array of integers
    two_l4: array of integers
    two_l5: array of integers
    two_l6: array of integers
    two_l7: array of integers
    two_l8: array of integers
    two_l9: array of integers
        Since L1, ..., L9 should be integers or half integers, two_l1 (which
        means 2 x L1) should be all integers.

    Returns
    -------
    threej: array
        The value of 9J symbol with the same shape of the arguments.
    """
    (
        two_la,
        two_lb,
        two_lc,
        two_ld,
        two_le,
        two_lf,
        two_lg,
        two_lh,
        two_lj,
    ) = int_broadcast(
        two_l1, two_l2, two_l3, two_l4, two_l5, two_l6, two_l7, two_l8, two_l9
    )

    # calculat summation directory from 6-j symbol
    x, sixj = drc6j(
        np.stack([two_lf, two_lb, two_la]),
        np.stack([two_lb, two_lf, two_lj]),
        np.stack([two_lc, two_le, two_lg]),
        np.stack([two_la, two_ld, two_lh]),
        np.stack([two_lj, two_lh, two_ld]),
    )
    neg_phase = (2 * x) % 4 - 1
    return np.sum(-neg_phase * (x + 1) * sixj[0] * sixj[1] * sixj[2], axis=-1)


def drc3jj(two_l2, two_l3, two_m2, two_m3, ignore_invalid=False):
    """
    Calculate Wigner's 3j symbol
    (  L1   L2 L3)
    (-M2-M3 M2 M3)
    for all the possible L1 values.

    Parameters
    ----------
    two_l2: array of integers, size (...)
    two_l3: array of integers, size (...)
    two_m2: array of integers, size (...)
    two_m3: array of integers, size (...)
        Since L2, ..., M3 should be integers or half integers, two_l1 (which
        means 2 x L1) should be all integers.

    Returns
    -------
    two_l1: 1d-np.ndarray of integer, shape (n, )
        The possible L1 values.
    threej: array, shape (..., n)
        The value of 3J symbol
    """
    two_l2, two_l3, two_m2, two_m3 = int_broadcast(two_l2, two_l3, two_m2, two_m3)

    if not ignore_invalid:
        if (two_l2 < 0).any() or (two_l3 < 0).any():
            raise ValueError("Some of l values are negative.")
        if (two_l2 < np.abs(two_m2)).any() or (two_l3 < np.abs(two_m3)).any():
            raise ValueError("Some of m values are larger than l.")
    else:
        valid = (two_l2 >= 0) * (two_l3 >= 0) * (two_l2 >= np.abs(two_m2)) * (two_l3 >= np.abs(two_m3))
        two_l2 = np.where(valid, two_l2, 0)
        two_l3 = np.where(valid, two_l3, 0)
        two_m2 = np.where(valid, two_m2, 0)
        two_m3 = np.where(valid, two_m3, 0)

    shape = two_l2.shape
    l1max = int(np.max(two_l2 + two_l3) + 1)

    thrcof, ier = _wigner.drc3jj_vec(
        two_l2=two_l2.ravel(),
        two_l3=two_l3.ravel(),
        two_m2=two_m2.ravel(),
        two_m3=two_m3.ravel(),
        nvec=two_l2.size,
        ndim=l1max,
    )
    thrcof = thrcof.reshape(shape + (l1max,))
    if ignore_invalid:
        thrcof = np.where(valid[..., np.newaxis], thrcof, 0)
    return np.arange(l1max), thrcof


def _drc3jj(two_l2, two_l3, two_m2, two_m3):
    """ scalar version of drc3jj """
    l1max = two_l2 + two_l3
    l1min = max(np.abs(two_l2 - two_l3), np.abs(two_m2 + two_m3))
    ndim = int((l1max - l1min + 2) / 2)

    l1min, l1max, thrcof, ier = _wigner.drc3jj_int(two_l2, two_l3, two_m2, two_m3, ndim)
    if ier == 1:
        raise ValueError("Either L2.LT.ABS(M2) or L3.LT.ABS(M3).")
    elif ier == 2:
        raise ValueError("Either L2+ABS(M2) or L3+ABS(M3) non-integer.")
    elif ier == 3:
        raise ValueError("L1MAX-L1MIN not an integer.")
    elif ier == 4:
        raise ValueError("L1MAX less than L1MIN.")
    elif ier == 5:
        raise ValueError("NDIM less than L1MAX-L1MIN+1.")

    return np.arange(l1min, l1min + ndim * 2, 2), thrcof


def drc6j(two_l2, two_l3, two_l4, two_l5, two_l6, ignore_invalid=False):
    """
    Calculate Wigner's 6j symbol
    (L1 L2 L3)
    (L4 L5 L6)
    for all the possible L1 values.

    Parameters
    ----------
    two_l2: array of integers, size (...)
    two_l3: array of integers, size (...)
    two_l4: array of integers, size (...)
    two_l5: array of integers, size (...)
    two_l6: array of integers, size (...)
        Since L2, ..., L6 should be integers or half integers, two_l2, ...
        (whichs 2 x L1) should be all integers.

    Returns
    -------
    two l1: 1d-np.ndarray of integer, shape (n, )
        The possible L1 values.
    threej: array, shape (..., n)
        The value of 3J symbol
    """
    two_l2, two_l3, two_l4, two_l5, two_l6 = int_broadcast(
        two_l2, two_l3, two_l4, two_l5, two_l6
    )

    if not ignore_invalid:
        if (
            (two_l2 < 0).any()
            or (two_l3 < 0).any()
            or (two_l4 < 0).any()
            or (two_l5 < 0).any()
            or (two_l6 < 0).any()
        ):
            raise ValueError("Some of l values are negative")
    else:
        valid = (two_l2 >= 0) * (two_l3 >= 0) * (two_l4 >= 0) * (two_l5 >= 0) * (two_l6 >= 0)
        two_l2 = np.where(valid, two_l2, 0)
        two_l3 = np.where(valid, two_l3, 0)
        two_l4 = np.where(valid, two_l4, 0)
        two_l5 = np.where(valid, two_l5, 0)
        two_l6 = np.where(valid, two_l6, 0)

    shape = two_l2.shape
    l1max = int(np.max(two_l2 + two_l3) + 1)

    sixcof, ier = _wigner.drc6j_vec(
        two_l2=two_l2.ravel(),
        two_l3=two_l3.ravel(),
        two_l4=two_l4.ravel(),
        two_l5=two_l5.ravel(),
        two_l6=two_l6.ravel(),
        nvec=two_l2.size,
        ndim=l1max,
    )
    sixcof = sixcof.reshape(shape + (l1max,))
    if ignore_invalid:
        sixcof = np.where(valid[..., np.newaxis], sixcof, 0)
    return np.arange(l1max), sixcof


def _drc6j(two_l2, two_l3, two_l4, two_l5, two_l6):
    """ scalar version of drc6j """
    two_l1min = max(abs(two_l2 - two_l3), abs(two_l5 - two_l6))
    two_l1max = min(two_l2 + two_l3, two_l5 + two_l6)
    ndim = int((two_l1max - two_l1min) / 2 + 1)

    l1min, l1max, sixcof, ier = _wigner.drc6j_int(
        two_l2, two_l3, two_l4, two_l5, two_l6, ndim
    )

    if ier == 1:
        raise ValueError("L2+L3+L5+L6 and L2+L4+L6 must be integers")
    elif ier == 2:
        raise ValueError("ABS(L2-L4).LE.L6.LE.L2+L4 must be satisfied.")
    elif ier == 3:
        raise ValueError("ABS(L4-L5).LE.L3.LE.L4+L5 must be satisfied.")
    elif ier == 4:
        raise ValueError(
            "L1MAX-L1MIN must be a non-negative integer, "
            "where L1MAX=MIN(L2+L3,L5+L6) and"
            "L1MIN=MAX(ABS(L2-L3),ABS(L5-L6))."
        )

    return np.arange(l1min, l1min + ndim * 2, 2), sixcof



import numpy as np
import pytest
from py3nj import wigner9j


rng = np.random.RandomState(0)
# precalculated 3j symbols. Input, result
NINE_J = (
    ((1, 1, 1, 1, 1, 1, 1, 1, 1), 0.0),
    #
    ((1, 2, 1, 1, 1, 1, 1, 1, 1), 1.0 / 18),
    #
    ((1, 2, 1, 2, 3, 1, 1, 2, 1), 1.0 / 5.0 * np.sqrt(1.0 / 30)),
)


def test_wigner9j():
    # scalar test
    for nine_j, expected in NINE_J:
        nine_j = (np.array(nine_j) * 2).astype(int)
        actual = wigner9j(*nine_j)
        print(nine_j)
        assert np.allclose(actual, expected)
    # test vector
    nine_j = (np.array([nine for nine, _ in NINE_J]).T * 2).astype(int)
    expected = np.array([value for _, value in NINE_J]).T
    actual = wigner9j(*nine_j)
    assert np.allclose(actual, expected)


def test_issue2():
    # https://github.com/fujiisoup/py3nj/issues/2
    actual = wigner9j(0, 0, 0, 2 * 1, 0, 2 * 1, 2 * 1, 0, 2 * 1)
    expected = 1.0 / 3.0
    assert np.allclose(actual, expected)



import numpy as np
import pytest
from py3nj import clebsch_gordan


rng = np.random.RandomState(0)

# reference values from
# https://en.wikipedia.org/wiki/Table_of_Clebsch%E2%80%93Gordan_coefficients


CG = (
    # j1 = 1/2, j2 = 1/2
    ((1, 1, 2, 1, 1, 2), 1),
    ((1, 1, 2, -1, -1, -2), 1),
    ((1, 1, 2, 1, -1, 0), np.sqrt(0.5)),
    ((1, 1, 2, -1, 1, 0), np.sqrt(0.5)),
    ((1, 1, 0, 1, -1, 0), np.sqrt(0.5)),
    ((1, 1, 0, -1, 1, 0), -np.sqrt(0.5)),
    # j1 = 1, j2 = 1/2
    ((2, 2, 4, 2, 0, 2), np.sqrt(1.0 / 2)),
)


def test_wigner3j_value():
    # scalar test
    for three_j, expected in CG:
        print(three_j)
        three_j = np.array(three_j)
        actual = clebsch_gordan(*three_j)
        assert np.allclose(actual, expected)
    # test vector
    three_j = np.array([thr for thr, _ in CG]).T
    expected = np.array([value for _, value in CG]).T
    actual = clebsch_gordan(*three_j)
    assert np.allclose(actual, expected)



import numpy as np
import pytest
from py3nj import wigner, wigner3j


rng = np.random.RandomState(0)
# precalculated 3j symbols. Input, result
THREE_J = (
    ((1, 1, 0, 0, 0, 0), -np.sqrt(1.0 / 3)),
    ((0, 1, 1, 0, 0, 0), -np.sqrt(1.0 / 3)),
    ((2, 2, 0, 0, 0, 0), np.sqrt(1.0 / 5)),
    ((2, 1, 1, 0, 0, 0), np.sqrt(2.0 / 15)),
    ((2, 1, 1, 0, 0, 0), np.sqrt(2.0 / 15)),
    ((0, 1, 1, 0, -1, 1), np.sqrt(1.0 / 3)),
    ((0.5, 0.5, 1.0, -0.5, 0.5, 0), np.sqrt(1.0 / 6.0)),
    ((0.5, 0.5, 1.0, 0.5, -0.5, 0), np.sqrt(1.0 / 6.0)),
    ((0.5, 0.5, 0.0, 0.5, -0.5, 0), np.sqrt(1.0 / 2.0)),
    ((0.5, 0.5, 0.0, -0.5, 0.5, 0), -np.sqrt(1.0 / 2.0)),
    ((1, 2, 3, 0, 2, -2), -np.sqrt(1.0 / 21)),
    ((1, 2, 3, 1, 2, -3), np.sqrt(1.0 / 7)),
    ((4, 0, 0, 0, 0, 0), 0),
)


def test_drc3jj():
    for three_j, result in THREE_J:
        three_j = (np.array(three_j) * 2).astype(int)
        l, thrcof = wigner._drc3jj(three_j[1], three_j[2], three_j[4], three_j[5])

        if three_j[0] in l:
            assert np.allclose(thrcof[l == three_j[0]], result)
        else:
            assert np.allclose(thrcof, 1)


def test_invalid():
    with pytest.raises(ValueError):
        l, thrcof = wigner._drc3jj(1, 3, 0, 0)

    with pytest.raises(ValueError):
        l, thrcof = wigner._drc3jj(1, 3, 4, 0)


@pytest.mark.parametrize("half_integer", [1, 0])
def test_drc3jj_vec(half_integer):
    n = 1000
    l2 = rng.randint(0, 10, size=n) * 2 + half_integer
    l3 = rng.randint(0, 10, size=n) * 2 + half_integer
    m2 = np.zeros(n, dtype=int)
    m3 = np.zeros(n, dtype=int)
    for i in range(n):
        if l2[i] > 0:
            m2[i] = rng.randint(-l2[i], l2[i] + 1)
        if l3[i] > 0:
            m3[i] = rng.randint(-l3[i], l3[i] + 1)

    l, actual = wigner.drc3jj(l2, l3, m2, m3)
    for _ in range(2):
        l, actual = wigner.drc3jj(l2, l3, m2, m3)

    for i in range(n):
        if (actual[i] != 0).any():
            l, expected = wigner._drc3jj(l2[i], l3[i], m2[i], m3[i])
            assert np.allclose(actual[i, l], expected)
        else:
            with pytest.raises(ValueError):
                # all zeros or invalid argument
                l, expected = wigner._drc3jj(l2[i], l3[i], m2[i], m3[i])
                if (expected == 0).all():
                    raise ValueError


def test_drc3jj_vec_err():
    l2 = (0, 1, 2)
    l3 = (0, 1, 2)
    m2 = (1, 0, 0)
    m3 = (0, 0, 0)
    with pytest.raises(ValueError):
        _, actual = wigner.drc3jj(l2, l3, m2, m3)


@pytest.mark.parametrize("half_integer", [1, 0])
def test_wigner3j(half_integer):
    n = 10000
    n2 = 10
    l2 = rng.randint(0, 40, size=n) * 2 + half_integer
    l3 = rng.randint(0, 40, size=n) * 2 + half_integer
    m2 = np.zeros(n, dtype=int)
    m3 = np.zeros(n, dtype=int)
    for i in range(n):
        if l2[i] > 0:
            m2[i] = rng.randint(-l2[i], l2[i] + 1)
        if l3[i] > 0:
            m3[i] = rng.randint(-l3[i], l3[i] + 1)

    l, expected_all = wigner.drc3jj(l2, l3, m2, m3)
    for _ in range(n2):
        lindex = rng.randint(0, 40, n) * 2 - 1 + half_integer
        two_l1 = l[lindex]
        two_m1 = -(m2 + m3)
        actual = wigner3j(two_l1, l2, l3, two_m1, m2, m3)
        expected = expected_all[np.arange(len(lindex)), lindex]
        assert np.allclose(actual, expected)


@pytest.mark.parametrize("ignore_invalid", [True, False])
def test_wigner3j_value(ignore_invalid):
    # scalar test
    for three_j, expected in THREE_J:
        three_j = (np.array(three_j) * 2).astype(int)
        actual = wigner3j(*three_j, ignore_invalid=ignore_invalid)
        assert np.allclose(actual, expected)
        # broadcast the first argument
        actual = wigner3j(
            np.array([three_j[0]])[:, np.newaxis],
            *three_j[1:],
            ignore_invalid=ignore_invalid
        )
        assert np.allclose(actual, expected)
        # further broadcasting
        actual = wigner3j(
            np.array([three_j[0]])[:, np.newaxis, np.newaxis], *three_j[1:]
        )
        assert np.allclose(actual, expected)

    # test vector
    three_j = (np.array([thr for thr, _ in THREE_J]).T * 2).astype(int)
    expected = np.array([value for _, value in THREE_J]).T
    actual = wigner3j(*three_j)
    assert np.allclose(actual, expected)


def test_wigner3j_ignore_invalid():
    N = np.arange(1, 2)
    K = np.arange(-2, 3)[:, np.newaxis]
    with pytest.raises(ValueError):
        # should raise an error
        wigner3j(N * 2, K * 2, 0, 0, 0, 0, ignore_invalid=False)
    # should compute
    wigner3j(N * 2, K * 2, 0, 0, 0, 0, ignore_invalid=True)


def test0d():
    for _ in range(100):
        l2 = rng.randint(0, 40) * 2
        l3 = rng.randint(0, 40) * 2
        m2 = 0
        if l2 > 0:
            m2 = rng.randint(-l2, l2 + 1)
        m3 = 0
        if l3 > 0:
            m3 = rng.randint(-l3, l3 + 1)
        l, actual = wigner.drc3jj(l2, l3, m2, m3)
        assert actual.ndim == 1

        if (actual != 0).any():
            l, expected = wigner._drc3jj(l2, l3, m2, m3)
            assert np.allclose(actual[l], expected)
        else:
            with pytest.raises(ValueError):
                # all zeros or invalid argument
                l, expected = wigner._drc3jj(l2, l3, m2, m3)
                if (expected == 0).all():
                    raise ValueError


def test2d():
    n = (10, 11)
    l2 = rng.randint(0, 10, size=n) * 2
    l3 = rng.randint(0, 10, size=n) * 2
    m2 = np.zeros(n, dtype=int)
    m3 = np.zeros(n, dtype=int)
    for i0 in range(n[0]):
        for i1 in range(n[1]):
            i = (i0, i1)
            if l2[i] > 0:
                m2[i] = rng.randint(-l2[i], l2[i] + 1)
            if l3[i] > 0:
                m3[i] = rng.randint(-l3[i], l3[i] + 1)

    l, actual = wigner.drc3jj(l2, l3, m2, m3)
    for i0 in range(n[0]):
        for i1 in range(n[1]):
            i = (i0, i1)
            if (actual[i] != 0).any():
                l, expected = wigner._drc3jj(l2[i], l3[i], m2[i], m3[i])
                assert np.allclose(actual[i][l], expected)
            else:
                with pytest.raises(ValueError):
                    # all zeros or invalid argument
                    l, expected = wigner._drc3jj(l2[i], l3[i], m2[i], m3[i])
                    if (expected == 0).all():
                        raise ValueError



import numpy as np
import pytest
from py3nj import wigner, wigner6j


rng = np.random.RandomState(0)
# precalculated 3j symbols. Input, result
SIX_J = (
    ((0, 0, 0, 0, 0, 0), 1.0),
    ((2, 2, 2, 2, 2, 2), -3.0 / 70),
    ((4, 2, 2, 2, 2, 2), 2.0 / 35),
    ((4, 3, 2, 2, 2, 2), -(1.0 / 7) * np.sqrt(1.0 / 2)),
    ((0.5, 0.5, 1, 0.5, 0.5, 1), 1.0 / 6),
    ((0.5, 1.5, 1, 0.5, 0.5, 1), -(1.0 / 3)),
)


def test_drc6j():
    for six_j, result in SIX_J:
        six_j = (np.array(six_j) * 2).astype(np.int32)
        l, sixcof = wigner._drc6j(*(six_j[1:]))
        assert six_j[0] in l
        assert np.allclose(sixcof[l == six_j[0]], result)


def test_drc6j_vec():
    n = 1000
    l2 = rng.randint(0, 10, size=n)
    l3 = rng.randint(0, 10, size=n)
    l4 = rng.randint(0, 10, size=n)
    l5 = rng.randint(0, 10, size=n)
    l6 = rng.randint(0, 10, size=n)

    l, actual = wigner.drc6j(l2, l3, l4, l5, l6)
    for _ in range(2):
        l, actual = wigner.drc6j(l2, l3, l4, l5, l6)

    for i in range(n):
        if (actual[i] != 0).any():
            l, expected = wigner._drc6j(l2[i], l3[i], l4[i], l5[i], l6[i])
            assert np.allclose(actual[i, l], expected)
        else:
            with pytest.raises(ValueError):
                # all zeros or invalid argument
                l, expected = wigner._drc6j(l2[i], l3[i], l4[i], l5[i], l6[i])
                if (expected == 0).all():
                    raise ValueError


def test_wigner6j():
    n = 10000
    n2 = 10
    l2 = rng.randint(0, 40, size=n)
    l3 = rng.randint(0, 40, size=n)
    l4 = rng.randint(0, 40, size=n)
    l5 = rng.randint(0, 40, size=n)
    l6 = rng.randint(0, 40, size=n)

    l, expected_all = wigner.drc6j(l2, l3, l4, l5, l6)
    for _ in range(n2):
        lindex = rng.randint(0, 40, n)
        two_l1 = l[lindex]
        actual = wigner6j(two_l1, l2, l3, l4, l5, l6)
        expected = expected_all[np.arange(len(lindex)), lindex]
        assert np.allclose(actual, expected)


@pytest.mark.parametrize("ignore_invalid", [True, False])
def test_wigner6j_value(ignore_invalid):
    # scalar test
    for six_j, expected in SIX_J:
        six_j = (np.array(six_j) * 2).astype(int)
        actual = wigner6j(*six_j, ignore_invalid=ignore_invalid)
        assert np.allclose(actual, expected)
    # test vector
    six_j = np.array([six for six, _ in SIX_J]).T * 2
    expected = np.array([value for _, value in SIX_J]).T
    actual = wigner6j(*six_j.astype(int), ignore_invalid=ignore_invalid)
    assert np.allclose(actual, expected)


def test_wigner6j_GH13():
    """
    Test for https://github.com/fujiisoup/py3nj/issues/13
    """
    N = np.arange(10)
    J = np.arange(11)[:, np.newaxis]
    # should not fail
    wigner6j(0, N * 2, N * 2, J * 2, 2, 2)


def test_wigner6j_ignore_invalid():
    N = np.arange(1, 2)
    K = np.arange(-2, 3)[:, np.newaxis]
    with pytest.raises(ValueError):
        # should raise an error
        wigner6j(N * 2, K * 2, 0, 0, 0, 0, ignore_invalid=False)
    # should compute
    wigner6j(N * 2, K * 2, 0, 0, 0, 0, ignore_invalid=True)
